"""
Copyright (c) Facebook, Inc. and its affiliates.

This source code is licensed under the MIT license found in the
LICENSE file in the root directory of this source tree.
"""

import torch
from torch import nn
from torch.nn import functional as F


class ConvBlockUnet(nn.Module):
    """
    A Convolutional Block that consists of two convolution layers each followed by
    instance normalization, LeakyReLU activation and dropout.
    """

    def __init__(self, in_chans: int, out_chans: int, drop_prob: float):
        """
        Args:
            in_chans: Number of channels in the input.
            out_chans: Number of channels in the output.
            drop_prob: Dropout probability.
        """
        super().__init__()

        self.in_chans = in_chans
        self.out_chans = out_chans
        self.drop_prob = drop_prob

        self.layers = nn.Sequential(
            nn.Conv2d(in_chans, out_chans, kernel_size=3, padding=1, bias=False),
            nn.InstanceNorm2d(out_chans),
            nn.LeakyReLU(negative_slope=0.2, inplace=True),
            nn.Dropout2d(drop_prob),
            nn.Conv2d(out_chans, out_chans, kernel_size=3, padding=1, bias=False),
            nn.InstanceNorm2d(out_chans),
            nn.LeakyReLU(negative_slope=0.2, inplace=True),
            nn.Dropout2d(drop_prob),
        )

    def forward(self, image: torch.Tensor) -> torch.Tensor:
        """
        Args:
            image: Input 4D tensor of shape `(N, in_chans, H, W)`.

        Returns:
            Output tensor of shape `(N, out_chans, H, W)`.
        """
        return self.layers(image)


class SPPblock(nn.Module):
    def __init__(self, in_chans):
        super().__init__()
        """
        Args:
            in_chans: Number of channels in the input.            
        """
        self.pool1 = nn.MaxPool2d(kernel_size=2, stride=2)
        self.pool2 = nn.MaxPool2d(kernel_size=3, stride=3)
        self.pool3 = nn.MaxPool2d(kernel_size=5, stride=5)
        self.pool4 = nn.MaxPool2d(kernel_size=6, stride=6)

        self.conv = nn.Conv2d(in_channels=in_chans, out_channels=1, kernel_size=1, padding=0)

    def forward(self, x):
        """
        Args:
            image: Input 4D tensor of shape `(N, in_chans, H, W)`.

        Returns:
            Output tensor of shape `(N, in_chans+4, H, W)`.
        """
        h, w = x.size(2), x.size(3)
        layer1 = F.interpolate(self.conv(self.pool1(x)), size=(h, w), mode='bilinear')
        layer2 = F.interpolate(self.conv(self.pool2(x)), size=(h, w), mode='bilinear')
        layer3 = F.interpolate(self.conv(self.pool3(x)), size=(h, w), mode='bilinear')
        layer4 = F.interpolate(self.conv(self.pool4(x)), size=(h, w), mode='bilinear')

        out = torch.cat([layer1, layer2, layer3, layer4], 1)

        return out


class ZoomInZoomOut(nn.Module):
    def __init__(self, in_chans, zoom_factor):
        """
        Args:
            in_chans: Number of channels in the input.
            zoom_factor: Number of zoom factor.
        """
        super().__init__()
        self.zoomin = nn.Conv2d(in_channels=in_chans, out_channels=in_chans, kernel_size=3, stride=zoom_factor, padding=0)
        self.conv = nn.Conv2d(in_channels=in_chans, out_channels=in_chans, kernel_size=1, padding=0)
        self.conv_zoomout = nn.Conv2d(in_channels=in_chans, out_channels=in_chans, kernel_size=1, padding=0,stride=1)

    def forward(self, x):
        """
        Args:
            image: Input 4D tensor of shape `(N, in_chans, H, W)`.

        Returns:
            Output tensor of shape `(N, in_chans, H, W)`.
        """
        h, w =  x.size(2), x.size(3)
        # self.layer1 = self.conv_zoomout(F.interpolate(self.conv(self.zoomin(x)), size=(h, w), mode='bilinear'))
        conv1 = self.conv(x)
        zoomed = self.conv_zoomout(F.interpolate(self.zoomin(conv1), size=(h, w), mode='bilinear'))

        out = torch.add(conv1,zoomed)
        return out


class MultiFineDetailPooling(nn.Module):
    def __init__(self, in_chans, out_chans, zoom_factor):
        """
        Args:
            in_chans: Number of channels in the input.
            out_chans: Number of channels in the output.            
        """
        super().__init__()
        self.zoominzoomout = ZoomInZoomOut(in_chans=in_chans, zoom_factor=zoom_factor)
        self.multipooling = SPPblock(in_chans=in_chans)
        self.conv = nn.Conv2d(in_channels=in_chans + 4 + (in_chans*0), out_channels=out_chans, kernel_size=3, stride=1, padding=1)

    def forward(self, x):
        """
        Args:
            image: Input 4D tensor of shape `(N, in_chans, H, W)`.

        Returns:
            Output tensor of shape `(N, out_chans, H, W)`.
        """       
        add_zoom = torch.add(x, self.zoominzoomout(x))
        add_multipooling = torch.cat([add_zoom, self.multipooling(x)], dim=1)
        return self.conv(add_multipooling)
        

class mlpednet(nn.Module):
    """
    PyTorch implementation of a U-Net model with\n
    - multipooling and zoom-in zoom-out\n
    - pixel shuffling instead of transpose convolution\n
    Version 2 use MultiFineDetailPooling for the concat part only
    """

    def __init__(
        self,
        in_chans: int,
        out_chans: int,
        chans: int = 32,
        num_pool_layers: int = 4,
        drop_prob: float = 0.0,
    ):
        """
        Args:
            in_chans: Number of channels in the input to the U-Net model.
            out_chans: Number of channels in the output to the U-Net model.
            chans: Number of output channels of the first convolution layer.
            num_pool_layers: Number of down-sampling and up-sampling layers.
            drop_prob: Dropout probability.
        """
        super().__init__()

        self.in_chans = in_chans
        self.out_chans = out_chans
        self.chans = chans
        self.num_pool_layers = num_pool_layers
        self.drop_prob = drop_prob

        self.down_sample_layers = nn.ModuleList([ConvBlockUnet(in_chans, chans, drop_prob)])
        self.multi_fine_detail_poling = nn.ModuleList()        

        ch = chans
        for _ in range(num_pool_layers - 1):
            self.down_sample_layers.append(ConvBlockUnet(ch, ch * 2, drop_prob))                    
            ch *= 2
        self.conv = ConvBlockUnet(ch, ch * 2, drop_prob)
        
        pool_ch = chans
        for _ in range(num_pool_layers):            
            self.multi_fine_detail_poling.append(MultiFineDetailPooling(pool_ch,pool_ch, 2)) 
            pool_ch *= 2                        

        self.up_conv = nn.ModuleList()
        self.up_pixelshuffle = nn.ModuleList()
        self.matchChan_conv = nn.ModuleList()
        for _ in range(num_pool_layers - 1):
            self.up_pixelshuffle.append(nn.PixelShuffle(2))
            self.matchChan_conv.append(nn.Conv2d(in_channels=ch//2 , out_channels=ch, kernel_size=1, padding=0))
            self.up_conv.append(ConvBlockUnet(ch*2 , ch, drop_prob))
            ch //= 2

        self.up_pixelshuffle.append(nn.PixelShuffle(2))
        self.matchChan_conv.append(nn.Conv2d(in_channels=ch//2 , out_channels=ch, kernel_size=1, padding=0))
        self.up_conv.append(nn.Sequential(
                ConvBlockUnet(ch*2, ch, drop_prob),
                nn.Conv2d(ch, self.out_chans, kernel_size=1, stride=1),
            )
        )
        

    def forward(self, image: torch.Tensor) -> torch.Tensor:
        """
        Args:
            image: Input 4D tensor of shape `(N, in_chans, H, W)`.

        Returns:
            Output tensor of shape `(N, out_chans, H, W)`.
        """
        stack = [] 
        output = image
        
        # apply down-sampling layers
        for layer, mfdp_layer in zip(self.down_sample_layers, self.multi_fine_detail_poling):
            output = layer(output)
            output_mfdp = mfdp_layer(output)            
            stack.append(output_mfdp)               
            
            output = F.avg_pool2d(output, kernel_size=2, stride=2, padding=0)            

        output = self.conv(output)
        
        # apply up-sampling layers
        for pixelshuffle_conv, conv, match_conv in zip(self.up_pixelshuffle, self.up_conv, self.matchChan_conv):            
            multifinedetailpooling_layer = stack.pop()               

            output = match_conv(pixelshuffle_conv(output))                   
            
            # reflect pad on the right/botton if needed to handle odd input dimensions
            padding = [0, 0, 0, 0]
            if output.shape[-1] != multifinedetailpooling_layer.shape[-1]:
                padding[1] = 1  # padding right
            if output.shape[-2] != multifinedetailpooling_layer.shape[-2]:
                padding[3] = 1  # padding bottom
            if torch.sum(torch.tensor(padding)) != 0:
                output = F.pad(output, padding, "reflect")

            output = torch.cat([output, multifinedetailpooling_layer], dim=1)
            output = conv(output)         
                  
        return output
